# in qsm_forward/__init__.py
from .qsm_forward import forward_convolution, signal_model, add_noise, poly_fit_shim_like, compute_phase_offset, resize, crop, kspace_crop
from .visualisation import show_image, show_histogram
__all__ = ['forward_convolution', 'signal_model', 'add_noise', 'poly_fit_shim_like', 'compute_phase_offset', 'resize', 'crop', 'kspace_crop', 'show_image', 'show_histogram']




import numpy as np
from dipy.denoise.gibbs import gibbs_removal
from nilearn.image import resample_img

def forward_convolution(chi):
    """
    Perform the forward convolution operation.

    This function performs the forward convolution step of the QSM simulation.

    Parameters
    ----------
    chi : numpy.ndarray
        The susceptibility distribution array.

    Returns
    -------
    numpy.ndarray
        The resulting magnetic field array after the forward convolution operation.

    """
    dims = np.array(chi.shape)
    D = _generate_3d_dipole_kernel(data_shape=dims, voxel_size=[1, 1, 1], b0_dir=[0, 0, 1])
    
    chitemp = np.ones(2 * dims) * chi[-1, -1, -1]
    chitemp[:dims[0], :dims[1], :dims[2]] = chi
    field = np.real(np.fft.ifftn(np.fft.fftn(chitemp) * D))
    field = field[:dims[0], :dims[1], :dims[2]]

    return field

def signal_model(field, B0=3, TR=1, TE=30e-3, flip_angle=90, phase_offset=0, R1=1, R2star=50, M0=1):
    """
    Compute the MRI signal based on the given parameters.

    Parameters
    ----------
    field : numpy.ndarray
        The magnetic field distribution.
    B0 : float, optional
        The main magnetic field strength. Default is 3.
    TR : float, optional
        The repetition time. Default is 1.
    TE : float, optional
        The echo time. Default is 30e-3.
    flip_angle : float, optional
        The flip angle in degrees. Default is 90.
    phase_offset : float, optional
        The phase offset. Default is 0.
    R1 : float or numpy.ndarray, optional
        The longitudinal relaxation rate. Can be a single value or a 3D numpy array. Default is 1.
    R2star : float or numpy.ndarray, optional
        The effective transverse relaxation rate. Can be a single value or a 3D numpy array. Default is 50.
    M0 : float or numpy.ndarray, optional
        The equilibrium magnetization. Can be a single value or a 3D numpy array. Default is 1.

    Returns
    -------
    numpy.ndarray
        The computed MRI signal.

    """

    sigHR = M0 * np.exp(1j * (2 * np.pi * field * B0 * 42.58 * TE + phase_offset)) * np.exp(-TE * R2star) \
        * (1 - np.exp(-TR * R1)) * np.sin(np.deg2rad(flip_angle)) / (1 - np.cos(np.deg2rad(flip_angle)) * np.exp(-TR * R1))
    sigHR[np.isnan(sigHR)] = 0

    return sigHR

def add_noise(sig, peak_snr=np.inf):
    """
    Add complex Gaussian noise to a signal.

    Parameters
    ----------
    sig : numpy.ndarray
        The input signal to which noise will be added.
    peak_snr : float, optional
        The peak signal-to-noise ratio, by default np.inf

    Returns
    -------
    numpy.ndarray
        The input signal with added noise.

    """

    noise = np.random.randn(*sig.shape) + 1j * np.random.randn(*sig.shape)
    sig_noisy = sig + (noise * np.max(np.abs(sig))) / peak_snr
    return sig_noisy


def poly_fit_shim_like(volume, brain_mask, order=2):
    """
    Perform a polynomial fit of the given order to a volume using a binary brain mask to select points.

    Parameters
    ----------
    volume : numpy.ndarray
        3D array representing the volume to fit.
    brain_mask : numpy.ndarray
        3D binary array. Must be the same shape as `volume`. A True value at a coordinate will 
        include that point in the fit.
    order : int, optional
        The order of the polynomial to fit. Must be 0, 1, or 2. Default is 2.

    Returns
    -------
    FIT3D : numpy.ndarray
        3D array representing the fitted volume.
    Residuals : numpy.ndarray
        3D array representing the residuals of the fit.
    b : numpy.ndarray
        1D array representing the coefficients of the fitted polynomial.

    Raises
    ------
    ValueError
        If `volume` and `brain_mask` shapes are not the same.
    """

    dim = volume.shape
    
    ## for volume fitting
    #brain_mask = np.ones(brain_mask.shape)
    indices = np.nonzero(brain_mask)
    x1, y1, z1 = indices
    R = volume[indices]
    b = None
    
    if len(indices[0]) > (3*order)**2:
        model = _create_model(x1, y1, z1, dim, order)
        b = np.linalg.pinv(model) @ R
        temp = R - model @ b
        del model, R
        
        indices = np.meshgrid(*[range(d) for d in dim], indexing='ij')
        x1, y1, z1 = [ind.flatten() for ind in indices]
        model = _create_model(x1, y1, z1, dim, order)
        
        Fit = model @ b
        del model
        
        FIT3D = Fit.reshape(dim)
        Residuals = (volume-FIT3D)
    else:
        FIT3D = np.zeros_like(volume)
        Residuals = (volume-FIT3D) * brain_mask
    
    return FIT3D, Residuals, b

def compute_phase_offset(M0, brain_mask, dims):
    """
    Compute the phase offset of an image.

    Parameters
    ----------
    M0 : numpy.ndarray
        The input image.
    brain_mask : numpy.ndarray
        A binary mask that indicates the region of the brain.
    dims : tuple of int
        The dimensions of the input image.

    Returns
    -------
    numpy.ndarray
        The phase offset of the input image.

    """

    c, w = _center_of_mass(M0)
    
    x, y, z = np.meshgrid(
        np.arange(1, dims[1]+1)-c[1],
        np.arange(1, dims[0]+1)-c[0],
        np.arange(1, dims[2]+1)-c[2]
    )
    
    temp = (x/w[1])**2 + (y/w[0])**2 + (z/w[2])**2
    
    max_temp = np.max(temp[brain_mask != 0])
    min_temp = np.min(temp[brain_mask != 0])
    
    phase_offset = -temp / (max_temp - min_temp) * np.pi

    return phase_offset


def resize(nii, voxel_size):
    """
    Resize a Nifti image.

    Parameters
    ----------
    nii : nibabel.nifti1.Nifti1Image
        The input Nifti image.
    voxel_size : list of float
        The desired voxel size after resizing.

    Returns
    -------
    nibabel.nifti1.Nifti1Image
        The resized Nifti image.

    """
    
    target_shape = np.array(np.round((np.array(nii.header.get_zooms()) / voxel_size) * np.array(nii.header.get_data_shape())), dtype=int)
    target_affine = np.diag(list(voxel_size) + [1])
    
    return resample_img(
        nii,
        target_affine=target_affine,
        target_shape=target_shape,
        interpolation='nearest'
    )


def crop(x, shape):
    """
    Crop a nD matrix around its center.

    Parameters
    ----------
    x : numpy.ndarray
        The input n-dimensional matrix.
    shape : tuple of int
        The desired shape after cropping.

    Returns
    -------
    numpy.ndarray
        The cropped matrix.

    """
        
    m = np.array(x.shape)
    s = np.array(shape)
    if s.size < m.size:
        s = np.concatenate((s, np.ones(m.size - s.size, dtype=int)))
    if np.array_equal(m, s):
        res = x
        return res
    idx = []
    for n in range(s.size):
        start = np.floor_divide(m[n], 2) + np.ceil(-s[n] / 2)
        end = np.floor_divide(m[n], 2) + np.ceil(s[n] / 2)
        idx.append(slice(int(start), int(end)))
    res = x[tuple(idx)]
    return res

def kspace_crop(volume, dims, scaling=False, gibbs_correction=True):
    """
    Crop a 3D volume in k-space and apply optional scaling and Gibbs ringing correction.

    Parameters
    ----------
    volume : numpy.ndarray
        The input 3D volume.
    dims : tuple of int
        The desired dimensions after cropping.
    scaling : bool, optional
        Whether to scale the cropped volume to maintain the total energy. Default is False.
    gibbs_correction : bool, optional
        Whether to apply Gibbs ringing correction. Default is True.

    Returns
    -------
    numpy.ndarray
        The cropped volume.

    """

    working_volume = np.fft.ifftn(np.fft.ifftshift(crop(np.fft.fftshift(np.fft.fftn(volume)), dims)))
    
    # gibbs correction is only needed for non-complex volumes
    if not np.iscomplexobj(volume):
        working_volume = np.real(working_volume)
        
        if gibbs_correction:
            working_volume = gibbs_removal(gibbs_removal(working_volume, slice_axis=2), slice_axis=1)

    if scaling:
        working_volume *= np.prod(dims) / np.prod(volume.shape)
    
    return working_volume


def _generate_3d_dipole_kernel(data_shape, voxel_size, b0_dir):
    """
    Generate a 3D dipole kernel.

    This function generates a 3D dipole kernel used in the forward convolution step of the QSM simulation.

    Parameters
    ----------
    data_shape : tuple of int
        The shape of the data array (nx, ny, nz).
    voxel_size : list of float
        The size of a voxel in each direction (dx, dy, dz).
    b0_dir : list of float
        The direction of the B0 field (B0x, B0y, B0z).

    Returns
    -------
    numpy.ndarray
        A 3D array representing the dipole kernel.

    """
    kx, ky, kz = np.meshgrid(
        np.arange(-data_shape[1], data_shape[1]),
        np.arange(-data_shape[0], data_shape[0]),
        np.arange(-data_shape[2], data_shape[2])
    )

    kx = kx / (2 * voxel_size[0] * np.max(np.abs(kx)))
    ky = ky / (2 * voxel_size[1] * np.max(np.abs(ky)))
    kz = kz / (2 * voxel_size[2] * np.max(np.abs(kz)))

    k2 = kx**2 + ky**2 + kz**2
    k2[k2 == 0] = np.finfo(float).eps
    D = np.fft.fftshift(1 / 3 - ((kx * b0_dir[0] + ky * b0_dir[1] + kz * b0_dir[2])**2 / k2))
    
    return D


def _center_of_mass(data):
    """
    Compute the center of mass of a 3D array.

    Parameters
    ----------
    data : numpy.ndarray
        The input 3D array.

    Returns
    -------
    tuple
        A tuple containing two arrays:
        1) The coordinates of the center of mass.
        2) The standard deviation along each axis.

    """

    data = np.abs(data)
    dims = np.shape(data)
    coord = np.zeros(len(dims))
    width = np.zeros(len(dims))

    for k in range(len(dims)):
        dimsvect = np.ones(len(dims), dtype=int)
        dimsvect[k] = dims[k]
        temp = np.multiply(data, np.reshape(np.arange(1, dims[k]+1), dimsvect))
        coord[k] = np.sum(temp)/np.sum(data)
        temp = np.multiply(data, np.power(np.reshape(np.arange(1, dims[k]+1), dimsvect)-coord[k], 2))
        width[k] = np.sqrt(np.sum(temp)/np.sum(data))

    return coord, width

def _create_model(x1, y1, z1, dim, order):
    """
    Creates a model based on x, y, z coordinates and the specified order.

    Parameters
    ----------
    x1, y1, z1 : numpy.ndarray
        1D arrays of the x, y, z coordinates respectively.
    dim : tuple of int
        The shape of the 3D space.
    order : int
        The order of the model to create. Must be 0, 1, or 2.

    Returns
    -------
    model : numpy.ndarray
        2D array where each row is the model for the corresponding point.

    Raises
    ------
    ValueError
        If order is not 0, 1, or 2.
    """
    Nsize = [1, 4, 10, 20, 35]
    N = Nsize[order+1]
    model = np.zeros((len(x1), N), dtype=float)
    
    # zeroth order
    if order >= 0:
        model[:, 0] = 1
    
    # first order
    if order >= 1:
        model[:, 1] = np.reshape(x1 - dim[0]/2, (len(x1),))
        model[:, 2] = np.reshape(y1 - dim[1]/2, (len(x1),))
        model[:, 3] = np.reshape(z1 - dim[2]/2, (len(x1),))
    
    # second order
    if order >= 2:
        model[:, 4] = model[:, 1] * model[:, 1] - model[:, 2] * model[:, 2] # siemens x^2 - y^2
        model[:, 5] = model[:, 1] * model[:, 2] # x^1 y^1 z^0 - siemens xy
        model[:, 6] = 2 * model[:, 3] * model[:, 3] - model[:, 2] * model[:, 2] - model[:, 1] * model[:, 1] # 2 z^2 - x^2 - y^2
        model[:, 7] = model[:, 2] * model[:, 3] # x^0 y^1 z^1 - siemens yz
        model[:, 8] = model[:, 3] * model[:, 1] # x^1 y^0 z^1 - siemens xz

    return model





import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import nibabel as nib
import torch
from scipy import ndimage

def get_slice(volume, dim=0, slice_idx=None):
    if slice_idx is None:
        slice_idx = volume.shape[dim] // 2
    idx_tuple = tuple(slice_idx if i == dim else slice(None) for i in range(len(volume.shape)))
    return volume[idx_tuple], slice_idx

def show_histogram(data, mask=None, title=None, dim=0, n_bins=100, n_ticks=10, n_rotations=0, vrange_hist=None, vrange_ylim=None, vrange_imshow=None, show_range=True, show_mean=True, show_stds=True, cmap='gray', units=None, precision=3, **args):
    # if str
    if isinstance(data, str): data = torch.tensor(nib.load(data).get_fdata())
    
    # if not tensor
    if not isinstance(data, torch.Tensor): data = torch.tensor(data)

    # get slice
    slice, slice_idx = get_slice(data, dim=dim) if data.dim() == 3 else data
    if mask is not None:
        mask_slice = get_slice(mask)[0] if mask.dim() == 3 else mask
    
    # rotate slice
    if n_rotations:
        slice = np.rot90(slice, k=n_rotations)
        if mask is not None:
            mask_slice = np.rot90(mask_slice, k=n_rotations)
    
    # superfigure
    fig, (ax1, ax2) = plt.subplots(1,2)
    fig.set_figwidth(15)
    if title: fig.suptitle(title)#fontsize=29

    # data subplot
    if vrange_imshow is None: vrange_imshow = (float(data.min()), float(data.max()))
    ax1.set_title(f"volume[{slice_idx},:,:]" if dim == 0 else (f"volume[:,{slice_idx},:]" if dim == 1 else f"volume[:,:,{slice_idx}]"))
    im = ax1.imshow(slice, vmin=vrange_imshow[0], vmax=vrange_imshow[1], cmap=cmap, **args)
    if mask is not None: ax1.imshow(mask_slice, cmap='jet', interpolation='none', alpha=1.0*(mask_slice>0))
    ax1.set_axis_off()
    plt.colorbar(im, ax=ax1)#, orientation='horizontal')#, pad=0.2)

    # choose bins and ticks
    if n_bins <= 20:
        n_ticks = n_bins

    # histogram subplot
    if vrange_hist is None: vrange_hist = (float(data.min()), float(data.max()))
    hist = torch.histc(data, bins=n_bins, min=vrange_hist[0], max=vrange_hist[1])
    ax2.set_title("Histogram")
    ax2.bar(range(len(hist)), hist, align='center', color='skyblue')
    if units: ax2.set_xlabel(units)
    if vrange_ylim: ax2.set_ylim(vrange_ylim[0], vrange_ylim[1])
    if mask is not None:
        hist2 = torch.histc(data*mask, bins=n_bins, min=vrange_hist[0], max=vrange_hist[1])
        hist2 *= (max(hist) / max(hist2)) * 0.1
        ax2.bar(range(len(hist2)), hist2, align='center', color='red')
    
    ticks = np.array(np.linspace(start=0, stop=len(hist), num=n_ticks, endpoint=False))
    labels = [round(float(vrange_hist[0] + i*((vrange_hist[1] - vrange_hist[0]) / n_ticks)), precision) for i in range(n_ticks)]
    ax2.set_xticks(ticks=ticks, labels=labels, rotation=45, ha='right')

    def get_closest_tick(val):
        return np.argmin(np.abs(np.array(actual_ticks)-np.array(val)))
    
    def get_tick_from_val(val):
        return np.interp(val, (actual_ticks[0], actual_ticks[-1]), (ticks[0], ticks[-1]))

    actual_ticks = np.linspace(start=vrange_hist[0], stop=vrange_hist[1], num=n_bins)#.tolist()
    trans = ax2.get_xaxis_transform()
    if show_range:
        l1 = ax2.axvline(x=get_tick_from_val(vrange_imshow[0]), color='b')
        l2 = ax2.axvline(x=get_tick_from_val(vrange_imshow[1]), color='b')
        #l1 = ax2.axvline(x=get_closest_tick(vrange_imshow[0]), color='b')
        #l2 = ax2.axvline(x=get_closest_tick(vrange_imshow[1]), color='b')
        plt.text(get_tick_from_val(vrange_imshow[0]), .85, f"{round(vrange_imshow[0],precision)}", transform=trans, horizontalalignment='left' if vrange_imshow[0] <= vrange_hist[0] else 'right')
        plt.text(get_tick_from_val(vrange_imshow[1]), .85, f"{round(vrange_imshow[1],precision)}", transform=trans, horizontalalignment='right' if vrange_imshow[1] >= vrange_hist[1] else 'left')
    if show_mean:
        l3 = ax2.axvline(x=get_tick_from_val(data.mean()), color='r')
        plt.text(get_tick_from_val(data.mean()), .9, f"{round(float(data.mean()),precision)}±{round(float(data.std()),precision)}", transform=trans, horizontalalignment='left')
    if show_stds:
        l4 = ax2.axvline(x=get_tick_from_val(data.mean() + data.std()), color='darkred')
        l5 = ax2.axvline(x=get_tick_from_val(data.mean() - data.std()), color='darkred')

    plt.show()
    plt.close()

def show_image(
        nii_path=None,
        nii_object=None,
        numpy_array=None,
        slice_idx=None,
        dim_idx=0,
        rotations=0,
        title=None,
        caption=None,
        cmap='gray',
        interpolation='nearest',
        cbar=False,
        cbar_title=None,
        cbar_ticks=2,
        discrete=False,
        **kwargs
    ):

    # get image data
    if nii_path is not None:
        img_data = nib.load(nii_path).get_fdata()
    elif nii_object is not None:
        img_data = nii_object.get_fdata()
    elif numpy_array is not None:
        img_data = numpy_array
    else:
        raise ValueError()
    
    # only works with 3d data
    assert(len(img_data.shape) == 3)
    
    # dim_idx must be valid
    assert(0 <= dim_idx <= len(img_data.shape) - 1)
    
    # slice_idx must be valid
    if slice_idx is not None:
        assert(0 <= slice_idx <= img_data.shape[dim_idx])
    else:
        slice_idx = img_data.shape[dim_idx] // 2

    # get slice data
    if dim_idx == 0:
        slice_data = img_data[slice_idx,:,:]
    elif dim_idx == 1:
        slice_data = img_data[:,slice_idx,:]
    elif dim_idx == 2:
        slice_data = img_data[:,:,slice_idx]

    # rotate slice data
    if rotations != 0:
        slice_data = np.rot90(m=slice_data, k=1)

    fig, ax = plt.subplots()

    if title: ax.set_title(title)

    # Define the colormap
    num_colors = len(np.unique(img_data))
    unique_vals = np.unique(img_data)

    if discrete:
        cmap = plt.get_cmap(cmap, num_colors)
        midpoints = np.diff(unique_vals) / 2 + unique_vals[:-1]
        boundaries = np.concatenate(([unique_vals[0] - 0.5], midpoints, [unique_vals[-1] + 0.5]))
        norm = matplotlib.colors.BoundaryNorm(boundaries, num_colors)
        im = ax.imshow(X=slice_data, cmap=cmap, interpolation=interpolation, norm=norm, **kwargs)
    else:
        im = ax.imshow(X=slice_data, cmap=cmap, interpolation=interpolation, **kwargs)
    
    ax.set_xticks([])
    ax.set_yticks([])

    if cbar:
        im_ratio = slice_data.shape[0]/slice_data.shape[1]
        if discrete:
            cbar = ax.figure.colorbar(im, ax=ax, ticks=sorted(np.round(np.unique(img_data), decimals=3)), fraction=0.046*im_ratio, pad=0.04)# shrink=0.88)
            #cbar = fig.colorbar(im, fraction=1.0)
            cbar.ax.set_yticklabels(sorted(np.round(np.unique(img_data), decimals=3)))
        else:
            if all(x in kwargs for x in ['vmin', 'vmax']):
                cbar = ax.figure.colorbar(im, ax=ax, ticks=np.linspace(kwargs['vmin'], kwargs['vmax'], cbar_ticks), fraction=0.046*im_ratio, pad=0.04)
            else:
                cbar = ax.figure.colorbar(im, ax=ax, ticks=np.linspace(np.min(img_data), np.max(img_data), cbar_ticks), fraction=0.046*im_ratio, pad=0.04)
            #cbar = plt.colorbar(im, shrink=0.88)
        if cbar_title:
            cbar.ax.set_ylabel(cbar_title, rotation=0, ha='left')

    if caption: ax.set_xlabel(caption)

    plt.show()
    plt.close()

def view_slices_3d(image_3d, slice_nbr, vmin, vmax, title=""):
    fig = plt.figure(figsize=(12, 6))
    plt.suptitle(title, fontsize=10)

    plt.subplot(131)
    plt.imshow(np.take(image_3d, slice_nbr, 2), vmin=vmin, vmax=vmax, cmap="gray")
    plt.axis('off')
    plt.title("Axial")

    plt.subplot(132)
    image_rot = ndimage.rotate(np.take(image_3d, slice_nbr, 1), 90)
    plt.imshow(image_rot, vmin=vmin, vmax=vmax, cmap="gray")
    plt.axis('off')
    plt.title("Coronal")

    plt.subplot(133)
    image_rot = ndimage.rotate(np.take(image_3d, slice_nbr, 0), 90)
    plt.imshow(image_rot, vmin=vmin, vmax=vmax, cmap="gray")
    plt.axis('off')
    plt.title("Sagittal")
    cbar = plt.colorbar(shrink=0.5)




from setuptools import setup, find_packages

setup(
    name='qsm_forward',
    version='0.2-2',
    packages=find_packages(),
    url='https://github.com/astewartau/qsm-forward-model',
    author='Ashley Stewart',
    author_email='a.stewart.au@gmail.com',
    description='A forward-model simulation for Quantitative Susceptibility Mapping',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'numpy',
        'matplotlib',
        'nibabel',
        'nilearn',
        'ipykernel',
        'scikit-image',
        'dipy',
        'torch'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
)
